@@ -1,5 +1,6 @@
 import logging
 from collections import defaultdict
+from collections import deque
 from copy import copy
 from dataclasses import field
 from datetime import date
@@ -10,14 +11,12 @@ from enum import Enum
 from functools import cache
 from typing import Any
 from typing import DefaultDict
-from typing import Deque
 from typing import Dict
 from typing import Generic
 from typing import Iterable
 from typing import List
 from typing import Optional
 from typing import Set
-from typing import Tuple
 from typing import Type
 from typing import TypeVar
 from typing import cast
@@ -40,6 +39,7 @@ from dipdup.enums import IndexStatus
 from dipdup.enums import IndexType
 from dipdup.enums import ReindexingReason
 from dipdup.enums import TokenStandard
+from dipdup.exceptions import FrameworkException
 from dipdup.utils import json_dumps_decimals
 
 ParameterType = TypeVar('ParameterType', bound=BaseModel)
@@ -71,7 +71,7 @@ class OperationData:
     status: str
     has_internals: Optional[bool]
     storage: Any
-    diffs: Tuple[Dict[str, Any], ...] = field(default_factory=tuple)
+    diffs: tuple[Dict[str, Any], ...] = field(default_factory=tuple)
     block: Optional[str] = None
     sender_alias: Optional[str] = None
     nonce: Optional[int] = None
@@ -83,9 +83,11 @@ class OperationData:
     originated_contract_alias: Optional[str] = None
     originated_contract_type_hash: Optional[int] = None
     originated_contract_code_hash: Optional[int] = None
-    originated_contract_tzips: Optional[Tuple[str, ...]] = None
+    originated_contract_tzips: Optional[tuple[str, ...]] = None
     delegate_address: Optional[str] = None
     delegate_alias: Optional[str] = None
+    target_code_hash: Optional[int] = None
+    sender_code_hash: Optional[int] = None
 
 
 @dataclass
@@ -280,13 +282,13 @@ class VersionedTransaction:
 # NOTE: Overwritten by TransactionManager.register()
 def get_transaction() -> Optional[VersionedTransaction]:
     """Get metadata of currently opened versioned transaction if any"""
-    raise RuntimeError('TransactionManager is not registered')
+    raise FrameworkException('TransactionManager is not registered')
 
 
 # NOTE: Overwritten by TransactionManager.register()
-def get_pending_updates() -> Deque['ModelUpdate']:
+def get_pending_updates() -> deque['ModelUpdate']:
     """Get pending model updates queue"""
-    raise RuntimeError('TransactionManager is not registered')
+    raise FrameworkException('TransactionManager is not registered')
 
 
 class ModelUpdateAction(Enum):
@@ -374,7 +376,7 @@ class ModelUpdate(TortoiseModel):
                 elif isinstance(field_, fields.TimeField):
                     data[key] = time.fromisoformat(value)
 
-                # TODO: There may be more non-JSON-deserializable fields
+                # NOTE: There are possibly more non-JSON-deserializable fields.
 
         _logger.debug(
             'Reverting %s(%s) %s: %s',
@@ -404,7 +406,7 @@ class UpdateQuery(TortoiseUpdateQuery):
         annotations: Dict[str, Any],
         custom_filters: Dict[str, Dict[str, Any]],
         limit: Optional[int],
-        orderings: List[Tuple[str, str]],
+        orderings: List[tuple[str, str]],
         filter_queryset: TortoiseQuerySet,  # type: ignore[type-arg]
     ) -> None:
         super().__init__(
@@ -443,7 +445,7 @@ class DeleteQuery(TortoiseDeleteQuery):
         annotations: Dict[str, Any],
         custom_filters: Dict[str, Dict[str, Any]],
         limit: Optional[int],
-        orderings: List[Tuple[str, str]],
+        orderings: List[tuple[str, str]],
         filter_queryset: TortoiseQuerySet,  # type: ignore[type-arg]
     ) -> None:
         super().__init__(model, db, q_objects, annotations, custom_filters, limit, orderings)
